#include "PMsg_p.hpp"
#include "assert.h"

//==============================================================================

/* This is the MPI-aware class derived from Msg_p (which is MPI-agnostic).
It's not my finest. Ultimately we should be able to send to sets of ranks, and
lots of other fiddly stuff, but at the moment I'm just shoving functins in as
and when I come across a need.
*/

//==============================================================================

PMsg_p::PMsg_p(MPI_Comm c)
// constructor from empty
{
tag  = 0;                              // No tag
src  = -1;                             // UNDEF in the derived class, but the
tgt  = -1;                             // base is pure....
for(int i=0;i<Z_FIELDS;i++)subkey[i]=0x00;// No labels - yet...  
comm=c;
}
PMsg_p::PMsg_p(unsigned char * pb,int l,MPI_Comm c)
// constructor from byte stream
{
Load(pb,l);
comm=c;
}
PMsg_p::PMsg_p(PMsg_p & r)
// Copy constructor. By far the easiest way to do this is simply stream the
// source and unstream to the target....
{
Load(r.Stream(),r.Length());
comm=r.comm;
}
PMsg_p::~PMsg_p(){Clear();}

PMsg_p::Msg_p_t::Msg_p_t():data1(), data2()
{

}

//------------------------------------------------------------------------------

void PMsg_p::Send()
{
Send(Tgt());
}

//------------------------------------------------------------------------------

void PMsg_p::Bcast()
// I don't *know* that MPI_Bcast doesn't work.
// BUT I *do* know that this does.
{
if (comm==MPI_COMM_NULL) return;   // no communicator associated with this message?
int flag, sflag;
MPI_Finalized(&flag);
if (flag) return;   // MPI closed down already?
int Usize,Urank;
MPI_Comm_test_inter(comm,&flag); // sending over an intercomm?
if (flag)
{
   MPI_Comm_remote_size(comm,&Usize);  // MPI remote Universe size
   MPI_Comm_rank(MPI_COMM_WORLD,&Urank);  // My place in my local Universe
}
else
{
   MPI_Comm_size(comm,&Usize);  // MPI Universe size
   MPI_Comm_rank(comm,&Urank);  // My place within it
}
Src(Urank);
unsigned char * bs = Stream();         // Turn the message into a unified stream
unsigned len = Length();               // Length of stream
for (int i=0;i<Usize;i++) {            // Send a copy everywhere
  if (!flag && (i==Urank)) continue;   // ...except to myself
  Tgt(i);                              // Load target rank
  MPI_Request request;
  MPI_Ibsend(bs,len,MPI_CHAR,i,tag,comm,&request);
  MPI_Status status;
  do MPI_Test(&request,&sflag,&status);
  while (sflag==0);
}
}

//------------------------------------------------------------------------------

void PMsg_p::Clear()
{
// get rid of everything in the data structures.
M.data0 = 0;
M.data1.clear();
M.data2.clear();
vm.clear();
}

//------------------------------------------------------------------------------

int * PMsg_p::Get(int k)
// getter for the integer field. Technically we could eliminate the index k,
// but this simplifies things by keeping it most similar to the more-general
// PMsg_p class
{
if (k) return 0;
else return &M.data0; 
}
  
void PMsg_p::Get(int k,string & s)  // SAME ADR SPC AS CHAR
{
if (k == 1) s = M.data1;
else if (k== 2) s = M.data2;
else s = string();
}

//------------------------------------------------------------------------------

unsigned PMsg_p::Key()
{
unsigned ans = 0;
for (unsigned i=0;i<Z_FIELDS;i++)
  ans = (ans|subkey[i])<<(i==Z_FIELDS-1 ? 0 : BITSPERBYTE);
return ans;
}

//------------------------------------------------------------------------------

void PMsg_p::Key(unsigned char L0,unsigned char L1,unsigned char L2,unsigned char L3)
{
subkey[0] = L0;
subkey[1] = L1;
subkey[2] = L2;
subkey[3] = L3;
}

//------------------------------------------------------------------------------

unsigned PMsg_p::KEY(unsigned char L0,unsigned char L1,unsigned char L2,unsigned char L3)
{
unsigned ans = 0;
ans |= L0; ans <<= BITSPERBYTE;
ans |= L1; ans <<= BITSPERBYTE;
ans |= L2; ans <<= BITSPERBYTE;
ans |= L3;
return ans;
}

//------------------------------------------------------------------------------

int PMsg_p::Length()
// Retrieve length of streamed message
{
if (vm.empty()) Stream();
return (int)vm.size();
}

//------------------------------------------------------------------------------

void PMsg_p::Load(unsigned char * tm, int len)
// This routine knows nothing of types; *everything* is a byte string.
{
                                       // Copy the data somewhere safe
for(int i=0;i<len;i++) vm.push_back(tm[i]);
int nfc = 0;                           // Next free cell in byte stream
int msize;
pullB(nfc, msize);                     // Message size in bytes
pullB(nfc,tag);                        // Unload the tag
pullB(nfc,src);                        // Unload the source
pullB(nfc,tgt);                        // Unload the target
for(unsigned i=0;i<Z_FIELDS;i++) pullB(nfc, subkey[i]);
pullB(nfc,M.data0);
pullB(nfc,M.data1);
pullB(nfc,M.data2);
assert(msize==nfc);
}

//------------------------------------------------------------------------------

void PMsg_p::pullB(int & nfc, int & i)
{
int x = * new(&vm[nfc]) int;
nfc += sizeof(int);
i = x;
}

//------------------------------------------------------------------------------

void PMsg_p::pullB(int & nfc, unsigned & u)
{
unsigned x = * new(&vm[nfc]) unsigned;
nfc += sizeof(unsigned);
u = x;
}

//------------------------------------------------------------------------------

void PMsg_p::pullB(int & nfc, string & s)
{
s = string(static_cast<char*>(static_cast<void*>(&vm[nfc])));
nfc += (s.size()+1);
}

//------------------------------------------------------------------------------

void PMsg_p::pushB(vector<unsigned char> & v,int x)
// Routine to break up the variable x and shove it into the byte vector v
{
const int s = sizeof(int);                 // Size of x in bytes
unsigned char buff[s];                              // Can do this 'cos s is compile-time
new(buff) int(x);                          // Write x as a byte stream
for(int i=0;i<s;i++) v.push_back(buff[i]); // Save it
}

//------------------------------------------------------------------------------

void PMsg_p::pushB(vector<unsigned char> & v,unsigned x)
// Routine to break up the variable x and shove it into the byte vector v
{
const int s = sizeof(unsigned);            // Size of x in bytes
unsigned char buff[s];                              // Can do this 'cos s is compile-time
new(buff) unsigned(x);                     // Write x as a byte stream
for(int i=0;i<s;i++) v.push_back(buff[i]); // Save it
}

//------------------------------------------------------------------------------

void PMsg_p::pushB(vector<unsigned char> & v,char x)
// Routine to break up the variable x and shove it into the byte vector v
{
unsigned char buff[1];                 // Can do this 'cos s is compile-time
new(buff) char(x);                     // Write x as a byte stream
v.push_back(buff[0]);                  // Save it
}

//------------------------------------------------------------------------------

void PMsg_p::Put(int k,int i)
// very simple setter for the integer member
{
if (k) return;
M.data0 = i;
}

//------------------------------------------------------------------------------

void PMsg_p::Put(int k,string * s)
// and a similarly simple one for string members
{
if (k == 1) M.data1 = *s;
else if (k == 2) M.data2 = *s;
return;
}
  
//------------------------------------------------------------------------------

void PMsg_p::Send(int dest)
{
if (comm==MPI_COMM_NULL) return;   // no communicator associated with this message?
int flag,Usize;
if (MPI_Finalized(&flag)!=0) return;          // MPI closed down already?
MPI_Comm_test_inter(comm,&flag);              // Sending over an intercomm?
if (flag) MPI_Comm_remote_size(comm,&Usize);  // MPI remote Universe size
else MPI_Comm_size(comm,&Usize);              // MPI Universe size
if (dest < 0 || dest >= Usize) return;        // Invalid rank?
Tgt(dest);                                    // In case it's not there
MPI_Request request;
MPI_Status status;

MPI_Ibsend(Stream(),Length(),MPI_CHAR,dest,tag,comm,&request);

int i=0;
do { i++; MPI_Test(&request,&flag,&status); } while (flag==0);
MPI_Wait(&request,&status);

}

//------------------------------------------------------------------------------

unsigned char * PMsg_p::Stream()
// Routine to take all the byte vector fragments in the data structure and
// bolt them together to form a single byte stream that can be piped through MPI
{
if (!vm.empty()) return &vm[0];
pushB(vm,0);                      // Placeholder for total message size
pushB(vm,tag);                    // Message tag
pushB(vm,src);                    // Message source
pushB(vm,tgt);                    // Message target
unsigned i;
for(i=0;i<Z_FIELDS;i++) pushB(vm,subkey[i]);
pushB(vm,M.data0);
for (i=0;i<=M.data1.size();i++) pushB(vm,M.data1.c_str()[i]);
for (i=0;i<=M.data2.size();i++) pushB(vm,M.data2.c_str()[i]); 

new(&vm[0]) int(vm.size());       // Write message length into placeholder

return &vm[0];
}


//==============================================================================

