
//------------------------------------------------------------------------------

#include "MPITest.h"
#include "CommonBase.h"
#include <stdio.h>
#include "PMsg_p.hpp"
#include "Pglobals.h"

//------------------------------------------------------------------------------

void * test_func(void * args)
// Simple test function in a thread to observe behaviour of message send/receive
// across threads. 
{
PMsg_p tmsg; // a message to send
MPITest* pM = static_cast<MPITest*>(args);
tmsg.Src(pM->Urank);
tmsg.comm = pM->Comms[0];
while (pM->msgcount[2] < 4); // wait for a series of messages from outside
tmsg.Key(Q::TEST,Q::M1);     // when ready send 4 replies
tmsg.Tgt(pM->Urank ? pM->Urank-1 : pM->Usize[0]-1);
printf("MPITest rank %d child thread sending its wave of messages\n",pM->Urank);
fflush(stdout);
for (unsigned k = 0;k < 4;k++)
{
    tmsg.Put(0,k);
    tmsg.Send();
}
// perform an all-to-all (to test MPI-initiated collective communications)
unsigned r;
for (r = 0; r < pM->Usize[0]; r++) pM->testbuf[r] = r+pM->Urank;
printf("MPITest rank %d child thread initiating an AlltoAll\n",pM->Urank);
fflush(stdout);
MPI_Alltoall(MPI_IN_PLACE,0,MPI_UNSIGNED,pM->testbuf,1,MPI_UNSIGNED,MPI_COMM_WORLD);
for (r = 0; r < pM->Usize[0]; r++) printf ("MPITest rank %d has value %d in child buffer for process rank %d\n",pM->Urank, pM->testbuf[r],r);
fflush(stdout);
while (pM->msgcount[3] < 4); // wait for some more messages 
tmsg.Clear();
tmsg.Key(Q::CMND,Q::EXIT);   // exit when done by sending a message to ourself
tmsg.Send(pM->Urank);
pthread_exit(NULL);
return NULL;
}

//==============================================================================

MPITest::MPITest(int argc,char * argv[],string d):
  CommonBase(argc,argv,d,string(__FILE__))
{
FnMapx.push_back(new FnMap_t);    // create a new event map in the derived class

                                       // Load the message map
(*FnMapx[0])[PMsg_p::KEY(Q::TEST,Q::M0,Q::N000,Q::N000)] = &MPITest::OnMsg0;
(*FnMapx[0])[PMsg_p::KEY(Q::TEST,Q::M1,Q::N000,Q::N000)] = &MPITest::OnMsgT;
(*FnMapx[0])[PMsg_p::KEY(Q::TEST,Q::MN,Q::N000,Q::N000)] = &MPITest::OnMsgN; 
(*FnMapx[0])[PMsg_p::KEY(Q::CMND,Q::EXIT,Q::N000,Q::N000)] = &MPITest::OnExit;

mainbuf = new unsigned[Usize[0]];
testbuf = new unsigned[Usize[0]];

MPI_Barrier(Comms[0]);
 
void* targs = this;
 
if(pthread_create(&test_thread,NULL,test_func,targs))
  fprintf(stdout,"Error creating test_thread\n");

//printf("Back in main thread\n"); fflush(stdout);
PMsg_p imsg;                     // initial set of 4 messages to self
imsg.Src(Urank);
imsg.comm = Comms[0];
imsg.Key(Q::TEST,Q::M0);     
imsg.Tgt(Urank); // sending to ourselves
printf("MPITest rank %d sending first wave of messages\n",Urank);
fflush(stdout);
for (unsigned h = 0; h < 4; h++) imsg.Send();

MPISpinner();                          // Spin on *all* messages: exit on DIE

printf("********* MPITest rank %d on the way out\n",Urank); fflush(stdout);
}

//------------------------------------------------------------------------------

MPITest::~MPITest()
{
delete[] mainbuf;
delete[] testbuf;
WALKVECTOR(FnMap_t*, FnMapx, F)
    delete *F;

//printf("********* RTC rank %d destructor\n",Urank); fflush(stdout);
}

//------------------------------------------------------------------------------

unsigned MPITest::OnExit(PMsg_p * Z, unsigned cIdx)
// Exit command received from child thread
{
printf("MPITest rank %d: exiting. Waiting for child thread to terminate\n",Urank);
fflush(stdout); 
pthread_join(test_thread,NULL);             // wait until the child has exited
// MPI_Barrier(Comms[0]);                      // barrier to test single-thread MPISpinner() interception
return CommonBase::OnExit(Z,cIdx);          // Drop to the base class exit handler
}

//------------------------------------------------------------------------------

unsigned MPITest::OnMsg0(PMsg_p * Z, unsigned cIdx)
// First message wave
{
printf("Rank %d: received message %d of 4 in the first wave\n",Urank,msgcount[0]);
fflush(stdout);
if (++msgcount[0] < 4) return 0; // wait until all 4 initial messages have been received
MPI_Barrier(Comms[0]);           // then enter a barrier for test purposes
int msgn = 2;                    // Put interface expects a pointer to datatype, and it can't be a constant
PMsg_p amsg;                     // now issue a set of 4 messages to a child thread
amsg.Src(Urank);
amsg.comm = Comms[0];
amsg.Key(Q::TEST,Q::MN);     
amsg.Tgt(Urank+1 == Usize[0] ? 0 : Urank+1); // in the next process
amsg.Put(0,msgn);
printf("MPITest rank %d sending the second wave of messages\n",Urank);
fflush(stdout);
for (unsigned i = 0; i < 4; i++) amsg.Send();
return 0;
}

//------------------------------------------------------------------------------

unsigned MPITest::OnMsgT(PMsg_p * Z, unsigned cIdx)
// Second (child thread) message wave
{
int* msgval;
if (msgval = Z->Get(0))
   printf("Rank %d: received message %d of 4 in the child wave, with value %d\n",Urank,msgcount[1],*msgval);
else printf("Error: rank %d did not get a value for message %d in the child wave\n",Urank,msgcount[1]); 
fflush(stdout);  
if (++msgcount[1] < 4) return 0; // wait to receive all messages from the child
// perform an all-to-all (to test MPI-initiated collective communications)
MPI_Request mreq;
MPI_Status mstat;
int mcomplete = 0;
unsigned m;
for (m = 0; m < Usize[0]; m++) mainbuf[m] = m+Urank;
printf("MPITest rank %d initiating a nonblocking AlltoAll\n",Urank);
fflush(stdout);
MPI_Ialltoall(MPI_IN_PLACE,0,MPI_UNSIGNED,mainbuf,1,MPI_UNSIGNED,Comms[0],&mreq);
do {MPI_Test(&mreq,&mcomplete,&mstat);} while (!mcomplete);
for (m = 0; m < Usize[0]; m++) printf ("MPITest rank %d has value %d in buffer for process rank %d\n",Urank, mainbuf[m],m);
fflush(stdout);
int msgn = 3;                    // Put interface expects a pointer to datatype, and it can't be a constant
PMsg_p bmsg;                     // next set of 4 messages to the child
bmsg.Src(Urank);
bmsg.comm = Comms[0];
bmsg.Key(Q::TEST,Q::MN);     
bmsg.Tgt(Urank+1 == Usize[0] ? 0 : Urank+1); // in the next process
bmsg.Put(0,msgn);
printf("MPITest rank %d sending the third wave of messages\n",Urank);
fflush(stdout);
for (unsigned j = 0; j < 4; j++) bmsg.Send();
return 0;
}

//------------------------------------------------------------------------------

unsigned MPITest::OnMsgN(PMsg_p * Z, unsigned cIdx)
// Remaining message receives for child thread
{
int* N;
if (N = Z->Get(0)) ++msgcount[*N];
else return 0;
printf("Rank %d received message %d in the %d wave\n",Urank, msgcount[*N]-1,*N);
fflush(stdout);
}


//==============================================================================



